import { Component, OnInit, DoCheck, OnDestroy, ViewChild, ViewChildren, QueryList } from "@angular/core";
import { ActivatedRoute } from '@angular/router';

import { FormulaireService } from "../../services/formulaire/formulaire.service";
import { InternationalisationService } from "../../services/internationalisation/internationalisation.service";
import { FieldSet } from "../../formulaire/fieldset";
import { FormulaireDefinition } from "../../formulaire/formulaire-definition";
import { CalculatorResultsComponent } from "../../components/calculator-results/calculator-results.component";
import { Observer } from "../../services/observer";
import { Subscription } from "rxjs/Subscription";
import { FieldSetComponent } from "../field-set/field-set.component";
import { BaseComponent } from "../base/base.component";
import { CalculatorNameComponent } from "./calc-name.component";
import { SelectEntry } from "../../formulaire/select-entry";
import { FormulaireElement } from "../../formulaire/formulaire-element";
import { FieldsetContainer } from "../../formulaire/fieldset-container";

@Component({
    selector: 'hydrocalc',
    templateUrl: "./calculator.component.html",
    styles: [`
    .button_compute_ok {
        height: 3em;
        width: 30%;
    }

    .button_compute_err {
        height: 3em;
        width: 30%;
        color: red
    }

    #close-button {
        font-size: 1.5em;
    }
    `
    ]
})
export class GenericCalculatorComponent extends BaseComponent implements OnInit, DoCheck, OnDestroy, Observer {
    /**
     * liste des FieldSetComponent
     */
    @ViewChildren(FieldSetComponent)
    private _fieldsetComponents: QueryList<FieldSetComponent>;

    /**
     * composant d'affichage des résultats
     */
    @ViewChild(CalculatorResultsComponent)
    private resultsComponent: CalculatorResultsComponent;

    /**
     * composant "nom de la calculette"
     */
    @ViewChild(CalculatorNameComponent)
    private _calculatorNameComponent: CalculatorNameComponent;

    /**
     * formulaire affiché
     */
    private _formulaire: FormulaireDefinition;

    private _subscription: Subscription; // pour souscrire aux changements d'URL envoyés par le routeur

    /**
     * true si on a cliqué sur le bouton de lancement de calcul
     */
    private _computeClicked: boolean;

    /**
     * flag de validité gloable de la saisie
     * NB : la validité du bouton "calculer" dépend de la validité de la saisie dans l'UI et de celle indiquée par le formulaire.
     * La validité de l'UI comprend la forme (pas de chaîne alpha dans les champs numériques, etc..).
     * La validité formulaire comprend le domaine de définition des valeurs saisies.
     */
    private _isUIValid: boolean = false;

    /**
     * flag disabled du bouton "calculer"
     */
    private isCalculateDisabled: boolean = true;

    constructor(private intlService: InternationalisationService,
        private formulaireService: FormulaireService,
        private route: ActivatedRoute) {
        super();
    }

    private get formElements(): FormulaireElement[] {
        if (this._formulaire == undefined)
            return [];
        return this._formulaire.formElements;
    }

    /**
     * détermine si un FormulaireElement est du type FieldSet
     */
    private isFieldset(fe): boolean {
        return fe instanceof FieldSet && !fe.isTemplate;
    }

    /**
     * détermine si un FormulaireElement est du type FieldsetContainer
     */
    private isFieldsetContainer(fe): boolean {
        return fe instanceof FieldsetContainer;
    }

    private get hasForm() {
        return this._formulaire != undefined;
    }

    private get hasResults() {
        if (this.hasForm)
            return this._formulaire.hasResults;
        return false;
    }

    private get uitextTitre() {
        if (this.hasForm)
            return this.formulaireService.getLocalisedTitleFromCalculatorType(this._formulaire.calculatorType);

        return undefined;
    }

    private get uitextCalculer() {
        return this.intlService.localizeText("INFO_CALCULATOR_CALCULER");
    }

    private get uitextCloseDialogTitle() {
        return this.intlService.localizeText("INFO_CLOSE_DIALOGUE_TITRE");
    }

    private get uitextCloseDialogText() {
        return this.intlService.localizeText("INFO_CLOSE_DIALOGUE_TEXT");
    }

    private get uitextCloseDialogYes() {
        return this.intlService.localizeText("INFO_OPTION_YES");
    }

    private get uitextCloseDialogNo() {
        return this.intlService.localizeText("INFO_OPTION_NO");
    }

    ngOnInit() {
        this.intlService.addObserver(this);
        this.formulaireService.addObserver(this);
        this.subscribeRouter();
        this.formulaireService.updateLocalisation();
    }

    ngDoCheck() {
        if (this._formulaire != undefined)
            this.isCalculateDisabled = !(this._formulaire.isValid && this._isUIValid);
    }

    ngOnDestroy() {
        this.unsubscribeRouter();
        this.formulaireService.removeObserver(this);
        this.intlService.removeObserver(this);
    }

    private subscribeRouter() {
        // récupération des paramètres passés avec la route (URL)
        this._subscription = this.route.params.subscribe(params => {
            const uid: number = params['uid'];
            this.formulaireService.setCurrentForm(+uid);
        });
    }

    private unsubscribeRouter() {
        this._subscription.unsubscribe();
    }

    /*
     * gestion des événements clic sur les radios :
     * envoi d'un message au composant parent
     * cf. https://angular.io/guide/component-interaction#parent-listens-for-child-event
     */
    private onRadioClick(info: string) {
        let tmp: string[] = info.split("_");
        let symbol: string = tmp[0]; // symbole du paramètre dont vient l'événement radio
        let option: string = tmp[1]; // nouvel état (radio)

        this._formulaire.resetRadiosAndResults(symbol, option);
    }

    private onCloseForm() {
        this.formulaireService.requestCloseForm(this._formulaire.uid);
    }

    /**
     * met à jour l'interface
     */
    //    private updateUI() {
    // this.appRef.tick();
    // this.changeDetectorRef.detectChanges();  // provoque une détection des changements dans les contrôles
    //}

    private doCompute() {
        this._formulaire.doCompute();
        this._computeClicked = true;
    }

    private onCalcResultsViewChecked() {
        if (this._computeClicked) {
            this._computeClicked = false;
            this.scrollToResults();
        }
    }

    private scrollToResults() {
        const element = document.getElementById("resultsComp");
        if (element)
            element.scrollIntoView();
    }

    private getFieldsetStyleDisplay(id: string) {
        let isDisplayed: boolean = this._formulaire.isDisplayed(id);
        return isDisplayed ? "block" : "none";
    }

    private setForm(f: FormulaireDefinition) {
        if (this._formulaire != undefined)
            this._formulaire.removeObserver(this);
        this._formulaire = f;
        if (this._formulaire != undefined)
            this._formulaire.addObserver(this);
    }

    // interface Observer

    update(sender: any, data: any): void {
        if (sender instanceof InternationalisationService) {
            this.formulaireService.updateLocalisation();
        }
        else if (sender instanceof FormulaireService) {
            switch (data["action"]) {
                case "currentFormChanged":
                    const uid: number = +data["formId"];
                    this.setForm(this.formulaireService.getFormulaireFromId(uid));
                    this.resultsComponent.formulaire = this._formulaire;
                    this._calculatorNameComponent.model = this._formulaire;
                    break;
            }
        }
        else if (sender instanceof FormulaireDefinition) {
            switch (data["action"]) {
                case "resultsUpdated":
                    const f = sender as FormulaireDefinition;
                    if (this._formulaire.uid == f.uid)
                        this.resultsComponent.updateView();
            }
        }
    }

    /**
     * réception d'un événement d'un select
     */
    private onSelectChanged(val: SelectEntry) {
        this._formulaire.resetResults();
        this._formulaire.applyDependencies();
    }

    /**
     * appelé après le 1er affichage du composant
     */
    protected afterFirstViewChecked() {
        this.updateUIValidity();
    }

    /**
     * calcul de la validité globale de la vue
     */
    private updateUIValidity() {
        if (this._fieldsetComponents != undefined)
            this._isUIValid = this._fieldsetComponents.reduce(
                // callback
                (
                    // accumulator (valeur précédente du résultat)
                    acc,
                    // currentValue (élément courant dans le tableau)
                    fieldset,
                    // currentIndex (indice courant dans le tableau)
                    currIndex,
                    // array (tableau parcouru)
                    array
                ) => {
                    return acc && fieldset.isValid;
                }
                // valeur initiale
                , true);
    }

    /**
     * réception d'un événement de validité d'un FieldSetComponent
     */
    private OnFieldsetValid() {
        this.updateUIValidity();
    }
}
