import { CalculatorType, ComputeNodeType, ParamsEquation, Nub, acSection, RegimeUniforme, MethodeResolution } from "jalhyd";
import { CourbeRemous, CourbeRemousParams, ParamDomainValue, Result, ResultElement } from "jalhyd";
import { ParamsSectionRectang, cSnRectang, ParamsSectionCirc, cSnCirc, ParamsSectionPuiss, cSnPuiss } from "jalhyd";
import { ConduiteDistrib, ConduiteDistribParams, LechaptCalmon, LechaptCalmonParams, ParamsSectionTrapez, cSnTrapez } from "jalhyd";
import { PabDimension, PabDimensionParams, PabPuissance, PabPuissanceParams } from "jalhyd";

import { ParamService } from "../services/param/param.service";
import { InternationalisationService } from "../services/internationalisation/internationalisation.service";
import { ApplicationSetupService } from "../services/app-setup/app-setup.service";
import { Field } from "./field";
import { NgParameter, ParamRadioConfig, ParamValueMode } from "./ngparam";
import { InputField } from "./input-field";
import { SelectField } from "./select-field";
import { FieldSet } from "./fieldset";
import { FieldsetContainer } from "./fieldset-container";
import { FormulaireElement } from "./formulaire-element";
import { FixedVarResults, GraphType } from "../results/fixed-var-results";
import { SectionResults } from "../results/section-results";
import { RemousResults } from "../results/remous-results";
import { StringMap } from "../stringmap";
import { Observable, Observer, IObservable } from "../services/observer";
import { TopFormulaireElementIterator } from "./form-iterator/top-element-iterator";
import { DeepFieldsetIterator } from "./form-iterator/deep-fieldset-iterator";
import { DeepFormulaireElementIterator } from "./form-iterator/deep-element-iterator";
import { FormulaireNode } from "./formulaire-node";

export class FormulaireDefinition extends FormulaireNode implements IObservable, Observer {
    /**
     * implémentation par délégation de IObservable
     */
    private _observable: Observable;

    /**
     * symbole du paramètre à calculer par défaut (cf config "idCal")
     */
    private _defaultCalculatedParam: string;

    /**
     * id du SelectField configurant le type de section
     */
    private _sectionSelectFieldId: string;

    /**
     * Type de noeud de calcul actuel de la section (si la calculette est basée sur une section et qu'un menu permet 
     * de sélectionner le type de section)
     */
    private _sectionNodeType: ComputeNodeType = undefined;

    /**
     * résultats fixes/variables
     */
    private _fixVarResults: FixedVarResults;

    /**
     * résultats de section paramétrée
     */
    private _sectionResults: SectionResults;

    /**
     * résultats de courbes de remous
     */
    private _remousResults: RemousResults;

    /**
     * nom de de la calculette
     */
    private _calculatorName: string;

    constructor(
        private _calcType: CalculatorType,
        private paramService: ParamService,
        private intlService: InternationalisationService,
        private appSetupService: ApplicationSetupService
    ) {
        super();
        this._observable = new Observable;
        this._fixVarResults = new FixedVarResults();
        this._sectionResults = new SectionResults();
        this._remousResults = new RemousResults();
    }

    public get calculatorType(): CalculatorType {
        return this._calcType;
    }

    public get fixVarResults() {
        return this._fixVarResults;
    }

    public get sectionResults() {
        return this._sectionResults;
    }

    public get remousResults() {
        return this._remousResults;
    }

    public get formElements(): FormulaireElement[] {
        return this.kids as FormulaireElement[];
    }

    private getFieldSet(id: string): FieldSet {
        for (let fs of this.allFieldsets) {
            if (fs.id == id)
                return fs;
        }
        return undefined;
    }

    private hasParameter(symbol: string): boolean {
        for (const p of this.allFormElements) {
            if (p instanceof NgParameter)
                if (p.symbol === symbol)
                    return true;
        }
        return false;
    }

    public getParamFromSymbol(symbol: string): NgParameter {
        for (const p of this.allFormElements) {
            if (p instanceof NgParameter)
                if (p.symbol === symbol)
                    return p;
        }
        return undefined;
    }

    public getParamFromState(st: ParamRadioConfig): NgParameter {
        for (const p of this.allFormElements) {
            if (p instanceof NgParameter)
                if (p.radioState == st)
                    return p;
        }
        return undefined;
    }

    private removePrefix(s: string, prefix: string): string {
        if (s.startsWith(prefix)) {
            let l = prefix.length;
            return s.substr(l, s.length - l);
        }
        return undefined;
    }

    private get hasSectionNodeTypeSelect(): boolean {
        return this._sectionSelectFieldId != undefined;
    }

    private updateSectionNodeType() {
        this._sectionNodeType = this.getNodeTypeFromSelectField();
    }

    private getNodeTypeFromSelectField(): ComputeNodeType {
        if (this.hasSectionNodeTypeSelect) {
            let ssf: SelectField = <SelectField>this.getFormulaireElementById(this._sectionSelectFieldId);
            return this.getNodeTypeFromSelectValue(ssf.getValue().value);
        }

        return undefined;
    }

    private getNodeTypeFromSelectValue(select_val: string): ComputeNodeType {
        let sect: string = this.removePrefix(select_val, this._sectionSelectFieldId + "_");
        switch (sect) {
            case "trapez":
                return ComputeNodeType.SectionTrapeze;

            case "rect":
                return ComputeNodeType.SectionRectangle;

            case "circ":
                return ComputeNodeType.SectionCercle;

            case "puiss":
                return ComputeNodeType.SectionPuissance;

            default:
                throw new Error(`getComputeNodeTypeFromSection() : section ${sect} non pris en charge`);
        }
    }

    /**
     * remet les radios de tous les paramètres à FIX sauf "me" et ceux (celui) à l'état "except"
     */
    private resetOtherRadio(me: NgParameter, except: ParamRadioConfig) {
        for (const p of this.allFormElements) {
            if (p instanceof NgParameter)
                if (p != me && p.radioState != except && p.radioConfig != ParamRadioConfig.FIX)
                    p.radioState = ParamRadioConfig.FIX;
        }
    }

    /**
     * modifie les boutons radio "fix", "var", "cal" de tous les paramètres
     * en fonction de la modification de l'état d'un des paramètres
     * @param symbol symbole du paramètre source
     * @param option nouvel état "fix", "var" ou "cal" du paramètre source
     */
    public resetRadiosAndResults(symbol: string, option: string) {
        let sourceParam = this.getParamFromSymbol(symbol);
        let oldState: ParamRadioConfig = sourceParam.radioState;
        let newState: ParamRadioConfig = ParamRadioConfig[option.toUpperCase()];

        switch (oldState) {
            case ParamRadioConfig.FIX:
                switch (newState) {
                    case ParamRadioConfig.VAR:
                        this.resetOtherRadio(sourceParam, ParamRadioConfig.CAL);
                        break;

                    case ParamRadioConfig.CAL:
                        this.resetOtherRadio(sourceParam, ParamRadioConfig.VAR);
                        break;
                }
                break;

            case ParamRadioConfig.VAR:
                switch (newState) {
                    case ParamRadioConfig.CAL:
                        this.resetOtherRadio(sourceParam, ParamRadioConfig.VAR);
                        break;
                }
                break;

            case ParamRadioConfig.CAL:
                switch (newState) {
                    case ParamRadioConfig.FIX:
                        this.setDefault();
                        break;

                    case ParamRadioConfig.VAR:
                        this.resetOtherRadio(sourceParam, ParamRadioConfig.CAL);
                        this.setDefault();
                        break;
                }
        }

        sourceParam.radioState = newState;

        // on vérifie qu'il y a au moins un paramètre "à calculer" et sinon, on prend le 1er qui est à "fixé"
        if (this.getParamFromState(ParamRadioConfig.CAL) == undefined) {
            let newCal: NgParameter = undefined;

            for (const p of this.allFormElements) {
                if (p instanceof NgParameter)
                    if (p.radioConfig == ParamRadioConfig.CAL && p.radioState == ParamRadioConfig.FIX && p != sourceParam) {
                        newCal = p;
                        break;
                    }
                if (newCal != undefined)
                    break;
            }

            if (newCal != undefined)
                newCal.radioState = ParamRadioConfig.CAL;
        }

        this.resetResults();
        this.applyDependencies();
    }


    public applyDependencies() {
        for (const fe of this.allFormElements)
            fe.applyDependencies();
    }

    public resetResults() {
        this._fixVarResults.reset();
        this._sectionResults.reset();
        this._remousResults.reset();
    }

    /**
     * met le paramètre par défaut à CAL
     */
    private setDefault() {
        let defaultParamCal = this.getParamFromSymbol(this._defaultCalculatedParam);
        defaultParamCal.radioState = ParamRadioConfig.CAL;
    }

    private getDisplayedInputParameters(): NgParameter[] {
        let res = [];
        for (let fs of this.allFieldsets)
            if (fs.isDisplayed)
                for (let p of fs.fields)
                    if (p instanceof NgParameter && p.isDisplayed)
                        res.push(p);
        return res;
    }

    private getNodeParameterValue(nodeType: ComputeNodeType, symbol: string): number {
        for (let fs of this.allFieldsets) {
            for (let p of fs.fields) {
                if (p instanceof NgParameter)
                    // if (p.computeNodeType == nodeType && (p.symbol === symbol || p.alias == symbol)) {
                    if (p.computeNodeType == nodeType && p.symbol === symbol) {
                        switch (p.radioState) {
                            case ParamRadioConfig.FIX:
                                return p.getValue();

                            case ParamRadioConfig.VAR:
                            case ParamRadioConfig.CAL:
                                return undefined;
                        }
                    }
            }
        }

        if (nodeType != ComputeNodeType.None)
            return this.getNodeParameterValue(ComputeNodeType.None, symbol);

        throw "Formulaire.getNodeParameterValue() : pas de paramètre " + symbol + " trouvé pour le noeud " + ComputeNodeType[nodeType] + "(" + nodeType + ")";
    }

    private getParameterValue(symbol: string): number {
        return this.getNodeParameterValue(ComputeNodeType.None, symbol);
    }

    private getSectionComputedParam(): { symbol: string, label: string } {
        let targetSelect: SelectField = <SelectField>this.getFieldById("select_target");
        let targetValue: string = targetSelect.getValue().value;

        let symbol: string = this.removePrefix(targetValue, "select_target_");

        let label = targetSelect.getLabel();
        return { symbol, label };
    }

    public getFormulaireElementById(id: string): FormulaireElement {
        const res = this.getFormulaireNodeById(id);
        if (res instanceof FormulaireElement)
            return res;
        return undefined;
    }

    public getFieldById(id: string): Field {
        let res = this.getFormulaireElementById(id);
        if (res instanceof Field)
            return res;
        return undefined;
    }

    private parse_fieldset(json: {}): FieldSet {
        const nt: string = json["nodeType"];
        let node_type: ComputeNodeType = nt == undefined ? ComputeNodeType.None : ComputeNodeType[nt];

        const res: FieldSet = new FieldSet(node_type, this, json["type"] === "fieldset_template");
        res.parseConfig(json, { "paramService": this.paramService, "appSetupService": this.appSetupService });
        return res;
    }

    private parse_template_container(json: {}, templates: { [key: string]: FieldSet }) {
        const fsc: FieldsetContainer = new FieldsetContainer(this);
        fsc.parseConfig(json, templates);
        this.formElements.push(fsc);
    }

    private getOption(json: {}, option: string): string {
        for (let conf_index in json) {
            let conf = json[conf_index];
            if (conf["type"] === "options")
                return conf[option];
        }

        return undefined;
    }

    public parseConfig(json: {}) {
        this._defaultCalculatedParam = undefined;
        const templates: { [key: string]: FieldSet } = {};

        for (let conf_index in json) {
            const conf = json[conf_index];
            const type: string = conf["type"];

            switch (type) {
                // field set
                case "fieldset":
                case "fieldset_template":
                    const fs: FieldSet = this.parse_fieldset(conf);
                    if (fs.isTemplate)
                        templates[fs.id] = fs;
                    else
                        this.kids.push(fs);
                    break;

                // options globales
                case "options":
                    // id du paramètre à calculer par défaut

                    this._defaultCalculatedParam = conf["idCal"];
                    if (this._defaultCalculatedParam != undefined) {
                        let p = this.getParamFromSymbol(this._defaultCalculatedParam);
                        p.isDefault = true;
                        p.radioState = ParamRadioConfig.CAL;
                    }

                    // id du SelectField configurant le type de section
                    this._sectionSelectFieldId = this.getOption(json, "sectionSelectId");
                    break;

                case "template_container":
                    this.parse_template_container(conf, templates);
                    break;

                default:
                    throw new Error(`type d'objet de calculette ${type} non pris en charge`);
            }
        }

        // si le formulaire a un menu pour le type de section, on s'abonne à la valeur du champ correspondant
        if (this.hasSectionNodeTypeSelect) {
            const f: Field = this.getFieldById(this._sectionSelectFieldId);
            if (f == undefined || !(f instanceof SelectField))
                throw new Error(`le champ ${this._sectionSelectFieldId} n'est pas du type SelectField`);
            const sectTypeField: SelectField = f as SelectField;
            sectTypeField.addObserver(this);
            this.updateSectionNodeType()
        }

        // logObject(this._fieldSets, "fieldsets");
        // logObject(this._dependencies, "dependences");

        this.parseDependencies(json);
    }

    public parseDependencies(json: {}) {
        for (let conf_index in json) {
            const conf = json[conf_index];
            const type: string = conf["type"];

            switch (type) {
                // field set
                case "fieldset":
                case "template_container":
                    for (const k of this.kids)
                        if (k.id == conf["id"]) {
                            k.parseDependencies(conf);
                            break;
                        }
                    break;

                case "fieldset_template":
                    for (const k of this.kids)
                        if (k instanceof FieldsetContainer)
                            k.parseDependencies(conf);
                    break;
            }
        }
    }

    private isMySectionType(nodeType: ComputeNodeType, strict: boolean) {
        let res: boolean = nodeType == this._sectionNodeType;
        if (strict)
            return res;

        return res || this._calcType == CalculatorType.SectionParametree
    }

    private addFixedResults(displaySymbol: boolean) {
        for (const p of this.getDisplayedInputParameters())
            if (p.radioState == ParamRadioConfig.FIX && p.symbol !== "Pr")
                this._fixVarResults.addFixedResultElement(p, p.getValue(), displaySymbol);
    }

    private addSectionFixedResult(val: ResultElement, label: string, drawLabel: string = undefined) {
        this._sectionResults.addResult(val, label, drawLabel);
    }

    private addSectionFixedResults(displaySymbol: boolean) {
        for (let p of this.getDisplayedInputParameters())
            if (p.radioState == ParamRadioConfig.FIX && this.isMySectionType(p.computeNodeType, false) && p.symbol !== "Pr")
                this._fixVarResults.addFixedResultElement(p, p.getValue(), displaySymbol);
    }

    private getSectionVariatedParameter(): NgParameter {
        let res: NgParameter = this.getParamFromState(ParamRadioConfig.VAR);
        if (res != undefined && this.isMySectionType(res.computeNodeType, false))
            return res;
        return undefined;
    }

    private doComputeSectionVar(varParam: NgParameter) {
        let computePrec: number = this.getParameterValue("Pr"); // précision de calcul
        let nDigits = -Math.log10(computePrec);

        this.addSectionFixedResults(false);

        let computedParam = this.getSectionComputedParam();

        this._fixVarResults.setVariableParamHeaderFromParameter(varParam, false);
        this._fixVarResults.setVariableResultHeader(computedParam["label"]);

        var np: [acSection, ParamsEquation] = this.getSectionNubAndParameters(this._sectionNodeType);
        let sect: acSection = np[0];
        let prms: ParamsEquation = np[1];

        this._sectionResults.section = sect;

        let min: number = +varParam.minValue;
        let max: number = +varParam.maxValue;
        let step: number = +varParam.stepValue;

        let yField: InputField = <InputField>this.getFormulaireElementById("Y");
        let Y: number = yField.getValue();

        let compSymbol = computedParam["symbol"];
        for (let val = min; val <= max; val += step) {
            prms[varParam.symbol].v = val;

            sect.Reset(true);
            let res = sect.Calc(compSymbol, Y);
            this._fixVarResults.addVarResultElement(val, res.result);
        }

        this._fixVarResults.graphTitle = computedParam.symbol + " = f( " + varParam.symbol + " )";
    }

    private doComputeSection() {
        let varParam = this.getSectionVariatedParameter();
        if (varParam != undefined) {
            this.doComputeSectionVar(varParam);
            return;
        }

        var np: [acSection, ParamsEquation] = this.getSectionNubAndParameters(this._sectionNodeType);

        let sect: acSection = np[0];
        let prms: ParamsEquation = np[1];

        this._sectionResults.section = sect;

        let computePrec: number = this.getNodeParameterValue(ComputeNodeType.None, "Pr"); // précision de calcul
        let nDigits = -Math.log10(computePrec);

        let Y = prms.map.Y.v; // tirant d'eau original (doit être fourni à acSection.Calc() sous peine d'être modifié par les appels successifs car c'est en même temps un paramètre et une variable temporaire)

        // charge spécifique
        let Hs = sect.Calc("Hs", Y);
        this.addSectionFixedResult(Hs.result, this.intlService.localizeText("INFO_GRANDEUR_HS"), "Hs");

        // charge critique
        let Hsc = sect.Calc("Hsc", Y);
        this.addSectionFixedResult(Hsc.result, this.intlService.localizeText("INFO_GRANDEUR_HSC"), "Hsc");

        // largeur au miroir
        let B = sect.Calc("B", Y);
        this.addSectionFixedResult(B.result, this.intlService.localizeText("INFO_GRANDEUR_B"));

        // périmètre hydraulique
        let P = sect.Calc("P", Y);
        this.addSectionFixedResult(P.result, this.intlService.localizeText("INFO_GRANDEUR_P"));

        // surface hydraulique
        let S = sect.Calc("S", Y);
        this.addSectionFixedResult(S.result, this.intlService.localizeText("INFO_GRANDEUR_S"));

        // rayon hydraulique
        let R = sect.Calc("R", Y);
        this.addSectionFixedResult(R.result, this.intlService.localizeText("INFO_GRANDEUR_R"));

        // vitesse moyenne
        let V = sect.Calc("V", Y);
        this.addSectionFixedResult(V.result, this.intlService.localizeText("INFO_GRANDEUR_V"));

        // nombre de Froude
        let Fr = sect.Calc("Fr", Y);
        this.addSectionFixedResult(Fr.result, this.intlService.localizeText("INFO_GRANDEUR_FR"), );

        // tirant d'eau critique
        let Yc = sect.Calc("Yc", Y);
        this.addSectionFixedResult(Yc.result, this.intlService.localizeText("INFO_GRANDEUR_YC"), "Yc");

        // tirant d'eau normal
        let Yn = sect.Calc("Yn", Y);
        this.addSectionFixedResult(Yn.result, this.intlService.localizeText("INFO_GRANDEUR_YN"), "Yn");

        // tirant d'eau fluvial
        let Yf = sect.Calc("Yf", Y);
        this.addSectionFixedResult(Yf.result, this.intlService.localizeText("INFO_GRANDEUR_YF"), "Yf");

        // tirant d'eau torrentiel
        let Yt = sect.Calc("Yt", Y);
        this.addSectionFixedResult(Yt.result, this.intlService.localizeText("INFO_GRANDEUR_YT"), "Yt");

        // tirant d'eau conjugué
        let Yco = sect.Calc("Yco", Y);
        this.addSectionFixedResult(Yco.result, this.intlService.localizeText("INFO_GRANDEUR_YCO"), "Yco");

        // perte de charge
        let J = sect.Calc("J", Y);
        this.addSectionFixedResult(J.result, this.intlService.localizeText("INFO_GRANDEUR_J"));

        // Variation linéaire de l'énergie spécifique
        let IJ = sect.Calc("I-J", Y);
        this.addSectionFixedResult(IJ.result, this.intlService.localizeText("INFO_GRANDEUR_I-J"));

        // impulsion hydraulique
        let Imp = sect.Calc("Imp", Y);
        this.addSectionFixedResult(Imp.result, this.intlService.localizeText("INFO_GRANDEUR_IMP"));

        // contrainte de cisaillement
        let Tau0 = sect.Calc("Tau0", Y);
        this.addSectionFixedResult(Tau0.result, this.intlService.localizeText("INFO_GRANDEUR_TAU0"));
    }

    private doComputeRemous() {
        var np: [acSection, ParamsEquation] = this.getSectionNubAndParameters(this._sectionNodeType, false);

        let sect: acSection = np[0];
        let prmSect: ParamsEquation = np[1];

        let Yamont: number = this.getNodeParameterValue(this._sectionNodeType, "Yamont"); // tirant amont
        let Yaval: number = this.getNodeParameterValue(this._sectionNodeType, "Yaval"); // tirant aval
        let Dx: number = this.getNodeParameterValue(this._sectionNodeType, "Dx"); // pas de discrétisation
        let Long: number = this.getNodeParameterValue(this._sectionNodeType, "Long"); // longueur du bief
        let If: number = this.getNodeParameterValue(this._sectionNodeType, "If"); // pente du fond
        let YB: number = this.getNodeParameterValue(this._sectionNodeType, "YB"); // hauteur de berge
        let Yn: Result = sect.Calc("Yn"); // hauteur normale
        let Yc: Result = sect.Calc("Yc"); // hauteur critique

        this._remousResults.penteFond = If;

        // méthode de résolution

        let msf: SelectField = <SelectField>this.getFormulaireElementById("select_resolution");
        let smeth: string = msf.getValue().value;
        let methRes: MethodeResolution;
        if (smeth == "select_resolution_trap")
            methRes = MethodeResolution.Trapezes;
        else if (smeth == "select_resolution_rk4")
            methRes = MethodeResolution.RungeKutta4;
        else if (smeth == "select_resolution_euler")
            methRes = MethodeResolution.EulerExplicite;
        else
            throw "GenericCalculatorComponent.doComputeRemous() : type de méthode de résolution '" + smeth + "' inconnu";

        // paramètre supplémentaire à calculer

        let dsf: SelectField = <SelectField>this.getFormulaireElementById("select_target");
        let extraSymbol: string = this.removePrefix(dsf.getValue().value, "select_target_");

        // calcul

        let prmCR: CourbeRemousParams = new CourbeRemousParams(sect, Yamont, Yaval, Long, Dx, methRes);
        let cr = new CourbeRemous(prmCR);
        let res: Result = cr.calculRemous(extraSymbol == "none" ? undefined : extraSymbol);

        // affichage du graphe

        this._remousResults.hauteurBerge = new ResultElement(YB);
        this._remousResults.hauteurNormale = Yn.result;
        this._remousResults.hauteurCritique = Yc.result;
        if (extraSymbol != "none") {
            this._remousResults.extraParamLabel = dsf.getValue().label;
            this._remousResults.extraGraph = ["Hs", "Hsc", "Yf", "Yt", "Yco"].indexOf(extraSymbol) == -1;
        }
        else
            this._remousResults.extraGraph = false;

        // résultats numériques

        this._remousResults.addResult(res);
    }

    private getVariatedParameter(): NgParameter {
        return this.getParamFromState(ParamRadioConfig.VAR);
    }

    private getComputedParameter(): NgParameter {
        return this.getParamFromState(ParamRadioConfig.CAL);
    }

    private getNubAndParameters(): [Nub, ParamsEquation] {
        switch (this.calculatorType) {
            case CalculatorType.ConduiteDistributrice:
                {
                    let Q: number = this.getParameterValue("Q"); // débit Q
                    let D: number = this.getParameterValue("D"); // diamètre D
                    let J: number = this.getParameterValue("J"); // perte de charge J
                    let Lg: number = this.getParameterValue("Lg"); // Longueur de la conduite Lg
                    let Nu: number = this.getParameterValue("Nu"); // viscosité dynamique 
                    let prms = new ConduiteDistribParams(Q, D, J, Lg, Nu);
                    let nub = new ConduiteDistrib(prms);
                    return [nub, prms];
                }

            case CalculatorType.LechaptCalmon:
                {
                    let Q: number = this.getParameterValue("Q"); // débit Q
                    let D: number = this.getParameterValue("D"); // diamètre D
                    let J: number = this.getParameterValue("J"); // perte de charge J
                    let Lg: number = this.getParameterValue("Lg"); // Longueur de la conduite Lg
                    let L: number = this.getParameterValue("L"); // paramètre de matériau 1
                    let M: number = this.getParameterValue("M"); // paramètre de matériau 2
                    let N: number = this.getParameterValue("N"); // paramètre de matériau 3
                    let prms = new LechaptCalmonParams(Q, D, J, Lg, L, M, N);
                    let nub = new LechaptCalmon(prms);
                    return [nub, prms];
                }

            case CalculatorType.PabDimensions:
                {
                    let L: number = this.getParameterValue("L"); // longueur L
                    let W: number = this.getParameterValue("W"); // largeur W
                    let Y: number = this.getParameterValue("Y"); // tirant d'eau Y
                    let V: number = this.getParameterValue("V"); // volume V
                    let prms = new PabDimensionParams(L, W, Y, V);
                    let nub = new PabDimension(prms); // pour initialiser la calculabilité des paramètres
                    return [nub, prms];
                }

            case CalculatorType.PabPuissance:
                {
                    let DH: number = this.getParameterValue("DH");  // Chute entre bassins
                    let Q: number = this.getParameterValue("Q"); // Débit 
                    let V: number = this.getParameterValue("V"); // volume V
                    let Pv: number = this.getParameterValue("Pv"); // puissance dissipée
                    let prms = new PabPuissanceParams(DH, Q, V, Pv);
                    let nub = new PabPuissance(prms); // pour initialiser la calculabilité des paramètres
                    return [nub, prms];
                }

            default:
                throw "FormulaireDefinition.getNubAndParameters() : valeur de CalculatorType " + this.calculatorType + " non implémentée"
        }
    }

    private getSectionNubAndParameters(nt: ComputeNodeType, getY: boolean = true): [acSection, ParamsEquation] {
        // bief
        let Ks = this.getParameterValue("Ks"); // Strickler
        let If: number = this.getParameterValue("If"); // Pente du fond
        let YB: number = this.getParameterValue("YB"); // Hauteur de berge

        // caractéristiques hydro
        let Q: number = this.getParameterValue("Q"); // débit Q
        if (getY)
            var Y: number = this.getParameterValue("Y"); // tirant d'eau
        else
            Y = undefined;

        let Prec = this.getParameterValue("Pr"); // précision calcul/affichage

        // ??
        // let YCL: number = f.getParameterValue("YCL"); // Condition limite en cote à l'amont ou à l'aval
        // let Dx: number = f.getParameterValue("Dx"); // Pas d'espace (positif en partant de l'aval, négatif en partant de l'amont)
        // let Long: number = f.getParameterValue("Long"); // Longueur du bief

        switch (nt) {
            case ComputeNodeType.SectionTrapeze:
                {
                    let LargeurFond = this.getNodeParameterValue(nt, "LargeurFond"); // Largeur au fond
                    let Fruit = this.getNodeParameterValue(nt, "Fruit"); // Fruit des berges
                    let prms = new ParamsSectionTrapez(LargeurFond, Fruit, Y, Ks, Q, If, Prec, YB);
                    let cn = new cSnTrapez(prms);
                    cn.newtonMaxIter = this.appSetupService.newtonMaxIter;
                    return [cn, prms];
                }

            case ComputeNodeType.SectionRectangle:
                {
                    let LargeurFond = this.getNodeParameterValue(nt, "LargeurBerge"); // Largeur au fond
                    let prms = new ParamsSectionRectang(Y, LargeurFond, Ks, Q, If, Prec, YB);
                    let cn = new cSnRectang(prms);
                    cn.newtonMaxIter = this.appSetupService.newtonMaxIter;
                    return [cn, prms];
                }

            case ComputeNodeType.SectionCercle:
                {
                    let D = this.getNodeParameterValue(nt, "D"); // Largeur au fond
                    let prms = new ParamsSectionCirc(D, Y, Ks, Q, If, Prec, YB);
                    let cn = new cSnCirc(prms);
                    cn.newtonMaxIter = this.appSetupService.newtonMaxIter;
                    return [cn, prms];
                }

            case ComputeNodeType.SectionPuissance:
                {
                    let k = this.getNodeParameterValue(nt, "k"); // coefficient
                    let LargeurBerge = this.getNodeParameterValue(nt, "LargeurBerge"); // Largeur au niveau des berges
                    let prms = new ParamsSectionPuiss(k, Y, LargeurBerge, Ks, Q, If, Prec, YB);
                    let cn = new cSnPuiss(prms);
                    cn.newtonMaxIter = this.appSetupService.newtonMaxIter;
                    return [cn, prms];
                }

            default:
                throw "FormulaireDefinition.getSectionNubAndParameters() : valeur de ComputeNodeType " + nt + " non implémentée"
        }
    }

    /**
     * lance le calcul d'un paramètre en déterminant une valeur initiale
     */
    private runNubCalc(nub: Nub, p: NgParameter, prec: number): Result {
        let init: number;
        switch (p.domain.domain) {
            case ParamDomainValue.ANY:
            case ParamDomainValue.POS_NULL:
                init = 0;
                break;

            case ParamDomainValue.INTERVAL:
                init = p.domain.minValue;
                break;

            case ParamDomainValue.NOT_NULL:
            case ParamDomainValue.POS:
                init = 1e-8;
                break;
        }
        if (prec == undefined)
            return nub.Calc(p.symbol, init);

        return nub.Calc(p.symbol, init, prec);
    }

    private doComputeFixedVar() {
        let np: [Nub, ParamsEquation];
        let nub: Nub;
        let prms: ParamsEquation;
        let rg: boolean = this.calculatorType == CalculatorType.RegimeUniforme;
        if (rg) {
            let snp: [acSection, ParamsEquation] = this.getSectionNubAndParameters(this._sectionNodeType);
            nub = new RegimeUniforme(snp[0]);
            prms = snp[1];
        }
        else {
            np = this.getNubAndParameters();
            nub = np[0];
            prms = np[1];
        }

        if (this.hasParameter("Pr"))
            var computePrec: number = this.getParameterValue("Pr"); // précision de calcul

        let computedParam: NgParameter = this.getComputedParameter();

        let varParam: NgParameter = this.getVariatedParameter();
        if (varParam == undefined) {
            // pas de paramètre à varier

            let res: Result = this.runNubCalc(nub, computedParam, computePrec);
            if (res.ok) {
                this.addFixedResults(!rg);
                this._fixVarResults.addFixedResultElement(computedParam, res.result, !rg);
            }
            else {
                this._fixVarResults.addLogMessages(res.log);
            }
        }
        else {
            // il y a un paramètre à varier

            this.addFixedResults(!rg);
            this._fixVarResults.setVariableParamHeaderFromParameter(varParam, !rg);
            this._fixVarResults.setVariableResultHeaderFromParameter(computedParam);

            switch (varParam.valueMode) {
                case ParamValueMode.MINMAX:
                    this._fixVarResults.graphType = GraphType.Scatter;

                    let min: number = +varParam.minValue;
                    let max: number = +varParam.maxValue;
                    let step: number = +varParam.stepValue;

                    for (let val = min; val <= max; val += step) {
                        prms[varParam.symbol].v = val;

                        let res: Result = this.runNubCalc(nub, computedParam, computePrec);
                        if (res.ok) {
                            this._fixVarResults.addVarResultElement(val, res.result);
                        }
                        else {
                            this._fixVarResults.addLogMessages(res.log);
                        }
                    }
                    break;

                case ParamValueMode.LISTE:
                    this._fixVarResults.graphType = GraphType.Histogram;

                    for (let val of varParam.valueList) {
                        prms[varParam.symbol].v = val;

                        let res: Result = this.runNubCalc(nub, computedParam, computePrec);
                        if (res.ok) {
                            this._fixVarResults.addVarResultElement(val, res.result);
                        }
                        else {
                            this._fixVarResults.addLogMessages(res.log);
                        }
                    }

                    break;
            }

            this._fixVarResults.graphTitle = computedParam.symbol + " = f( " + varParam.symbol + " )";
        }
    }

    public doCompute() {
        this.resetResults();

        switch (this.calculatorType) {
            case CalculatorType.SectionParametree:
                this.doComputeSection();
                break;

            case CalculatorType.CourbeRemous:
                this.doComputeRemous();
                break;

            default:
                this.doComputeFixedVar();
                break;
        }

        this.notifyObservers({
            "action": "resultsUpdated",
        });
    }

    public isDisplayed(id: string) {
        return this.getFormulaireElementById(id).isDisplayed;
    }

    public get hasFixVarResults(): boolean {
        return this._fixVarResults.hasResults;
    }

    public get hasSectionResults(): boolean {
        return this._sectionResults.hasResults;
    }

    public get hasRemousResults(): boolean {
        return this._remousResults.hasResults;
    }

    public get hasResults(): boolean {
        return this.hasFixVarResults || this.hasSectionResults || this.hasRemousResults;
    }

    public updateLocalisation(localisation: StringMap) {
        for (let fe of this.topFormElements)
            fe.updateLocalisation(localisation);

        if (this.hasResults)
            this.doCompute(); // pour mettre à jour la langue
    }

    public get isValid(): boolean {
        let res: boolean = true;
        for (let fs of this.allFieldsets)
            res = res && fs.isValid;
        return res;
    }

    public get calculatorName() {
        return this._calculatorName;
    }

    public set calculatorName(n: string) {
        this._calculatorName = n;
        this.notifyObservers({
            "action": "nameChanged",
            "name": n
        });
    }

    /**
     * itère sur tous les FieldSet
     */
    private get allFieldsets(): IterableIterator<FieldSet> {
        return new DeepFieldsetIterator(this);
    }

    /**
     * itère sur tous les FormulaireElement de 1er niveau
     */
    private get topFormElements(): IterableIterator<FormulaireElement> {
        return new TopFormulaireElementIterator(this);
    }

    /**
     * itère sur tous les FormulaireElement
     */
    private get allFormElements(): IterableIterator<FormulaireElement> {
        return new DeepFormulaireElementIterator(this);
    }

    /**
     * copie des membres
     */
    protected copyMembers(n: FormulaireNode) {
        throw new Error("FormulaireDefinition.copyMembers() non implémenté");
    }

    /**
     * crée une nouvelle instance
     */
    protected clone(): FormulaireNode {
        throw new Error("FormulaireDefinition.clone() non implémenté");
    }

    // interface Observer 

    update(sender: IObservable, data: any): void {
        if (sender instanceof SelectField) {
            this.updateSectionNodeType();
        }
    }

    // interface IObservable

    /**
     * ajoute un observateur à la liste
     */
    public addObserver(o: Observer) {
        this._observable.addObserver(o);
    }

    /**
     * supprime un observateur de la liste
     */
    public removeObserver(o: Observer) {
        this._observable.removeObserver(o);
    }

    /**
     * notifie un événement aux observateurs
     */
    public notifyObservers(data: any) {
        this._observable.notifyObservers(data, this);
    }
}
