import { ComputeNodeType } from "jalhyd";

import { FormulaireNode } from "./formulaire-node"
import { StringMap } from "../stringmap";
import { Dependency } from "./dependency/dependency";
import { DependencyCondition, DependencyConditionType } from "./dependency/dependency-condition";
import { ValueDependency } from "./dependency/value-dependency";
import { ValueDependencyCondition } from "./dependency/value-dependency-condition";
import { ExistenceDependency } from "./dependency/existence-dependency";
import { FormulaireDefinition } from "./formulaire-definition";

/** 
 * élément (enfant) du formulaire : fieldset, input, container, ...
 */
export abstract class FormulaireElement extends FormulaireNode {
    /**
     * type de noeud de calcul
     */
    private _nodeType: ComputeNodeType;

    /**
     * formulaire parent
     */
    private _parentForm: FormulaireDefinition;

    /**
     * affiché ?
     */
    private _isDisplayed: boolean;

    /**
     * étiquette affichée dans l'UI
     */
    private _label: string;

    protected _dependencies: Dependency[] = [];

    constructor(nodeType: ComputeNodeType, parentForm: FormulaireDefinition, isTmpl: boolean = false) {
        super(isTmpl);
        this._nodeType = nodeType;
        this._isDisplayed = true;
        this._parentForm = parentForm;
    }

    get computeNodeType(): ComputeNodeType {
        return this._nodeType;
    }

    get parentForm(): FormulaireDefinition {
        return this._parentForm;
    }

    get isDisplayed(): boolean {
        return this._isDisplayed;
    }

    get label(): string {
        return this._label;
    }

    public getKids(): FormulaireElement[] {
        return super.kids as FormulaireElement[];
    }

    private isNumber(s: string): boolean {
        return Number(s) != NaN;
    }

    private parse_existence_dependencies(json: {}) {
        for (let di in json) {
            let d = json[di];
            let masterField: FormulaireElement = this.parentForm.getFormulaireElementById(d["refid"]);
            if (masterField != undefined) {
                let rv = d["refvalue"];
                if (rv != undefined)
                    var mc: DependencyCondition = new ValueDependencyCondition(rv);
                else {
                    let cond = d["cond"];
                    if (cond != undefined) {
                        switch (cond) {
                            case "isvar":
                                var mc = new DependencyCondition(DependencyConditionType.IsVariable);
                                break;

                            case "isdisp":
                                var mc = new DependencyCondition(DependencyConditionType.IsDisplayed);
                                break;

                            default:
                                throw "Formulaire.parse_existence_dependencies() : type de condition '" + cond + "' non pris en charge";
                        }
                    }
                    else
                        throw "Formulaire.parse_existence_dependencies() : infos de dépendance manquantes/non prises en charge";
                }
                let dep = new ExistenceDependency(masterField, mc);
                this._dependencies.push(dep);
            }
            else
                throw new Error(`la dépendance d'existence de '${this.id}' fait référence à un élément inconnu '${d["refid"]}'`);
        }
    }

    public parseDependencies(json: {}) {
        const dep = json["dep_exist"];
        if (dep != undefined)
            this.parse_existence_dependencies(dep);
    }

    protected abstract verifyDependency(d: Dependency): boolean;

    public verifiesDependency(d: Dependency): boolean {
        if (d.masterCondition.type == DependencyConditionType.IsDisplayed)
            return this._isDisplayed;

        return this.verifyDependency(d);
    }

    private getDependencyFromMasterValue(v: any): Dependency {
        for (let d of this._dependencies)
            if (d.masterCondition.type == DependencyConditionType.HasValue) {
                let mv = (<ValueDependencyCondition>d.masterCondition).value;
                if (mv === v)
                    return d;
            }
        return undefined;
    }

    private prepareExistenceDependencies() {
        // si des FormulaireElement sont présents dans des ExistenceDependency, on met leur membre isDisplayed à false

        for (let d of this._dependencies)
            if (d instanceof ExistenceDependency) {
                this._isDisplayed = false;
                break;
            }
    }

    public applyDependencies() {
        this.prepareExistenceDependencies();

        for (let d of this._dependencies) {
            let master: FormulaireElement = d.masterElement;
            if (d instanceof ExistenceDependency) {
                let slave: FormulaireElement = this;
                slave._isDisplayed = slave._isDisplayed || master.verifiesDependency(d);
            }
            else if (d instanceof ValueDependency) {
                let vd = <ValueDependency>d;
                /*
                let master: HTMLElement = this.getHtmlElementFromId(d.masterElement.id);
                if (this.getHtmlElementValue(master) == vd.masterValue) {
                    let slave: HTMLElement = this.getHtmlElementFromId(d.slaveElement.id);
                    this.setHtmlElementValue(slave, vd.slaveValue);
                }
                */
                if (master.verifiesDependency(d)) {
                    let slave = this.parentForm.getFieldById(this.id);
                    if (this.isNumber(vd.slaveValue))
                        slave.setValue(+vd.slaveValue);
                    else
                        slave.setValue(vd.slaveValue);
                }
            }
        }

        for (const k of this.getKids())
            k.applyDependencies();
    }

    public printDependencies() {
        for (let d of this._dependencies)
            console.log(d.toString());
    }

    /**
     * copie les dépendances de l'élément "this" vers un autre
     * @param dest élément vers lequel copier les dépendances
     * @param root élément racine instancié si on copie des dépendances de template
     */
    public copyDependencies(dest: FormulaireElement, root: FormulaireElement) {
        if (dest.isTemplate || dest._dependencies.length != 0 || root.isTemplate || this.id !== dest.id)
            throw new Error("copyDependencies() : erreur de cohérence (1)");

        // copie des dépendances
        // à priori, elles ne sont qu'avec des FormulaireElement en dehors du FieldsetContainer
        // ou avec des FormulaireElement du même FieldSet
        // cad pas avec des FormulaireElement dans des FieldSet dans le même FieldsetContainer

        for (const d of this._dependencies) {
            // si l'élément esclave est un template
            if (this.isTemplate) {
                // élément maître original
                const ome: FormulaireElement = d.masterElement;
                if (!ome.isTemplate)
                    throw new Error("copyDependencies() : erreur de cohérence (2)"); // la dépendance doit être entre éléments templates

                // on recherche l'instanciation de l'élément maître original dans l'élement racine
                let nme = root.getFormulaireNodeById(ome.id) as FormulaireElement;
                if (nme === undefined)
                    throw new Error("copyDependencies() : correspondance non trouvée");

                // copie de la dépendance

                const nd = d.clone(nme);
                dest._dependencies.push(nd);
            }
            else
                // pas de template impliqué, on réutilise la dépendance    
                dest._dependencies.push(d);
        }

        for (const k of this.kids) {
            const destKid = dest.getFormulaireElementById(k.id);
            (k as FormulaireElement).copyDependencies(destKid, root);
        }
    }

    public updateLocalisation(loc: StringMap) {
        this._label = loc[this.id];
        if (this._label == undefined)
            console.log(`WARNING : pas de traduction pour l'id ${this.id}`);

        for (let f of this.getKids())
            f.updateLocalisation(loc);
    }

    /**
     * copie des membres
     */
    protected copyMembers(n: FormulaireElement) {
        super.copyTo(n);
        n._isDisplayed = this._isDisplayed;
        n._label = this._label;
    }

    public toString() {
        return "id:" + this.id + (this._isDisplayed ? " displayed" : " NOT displayed") + " label:" + this.label;
    }

    public getFormulaireElementById(id: string): FormulaireElement {
        return this.getFormulaireNodeById(id) as FormulaireElement;
    }
}
