import { ComputeNodeType } from "jalhyd";

import { Field } from "./field";
import { SelectEntry } from "./select-entry";
import { Dependency } from "./dependency/dependency";
import { DependencyConditionType } from "./dependency/dependency-condition";
import { ValueDependencyCondition } from "./dependency/value-dependency-condition";
import { StringMap } from "../stringmap";
import { IObservable, Observable, Observer } from "../services/observer";
import { FormulaireDefinition } from "./formulaire-definition";


export class SelectField extends Field implements IObservable {
    private _entries: SelectEntry[];

    private _selectedEntry: SelectEntry;

    private _observable: Observable;

    constructor(nodeType: ComputeNodeType, parentForm: FormulaireDefinition, isTmpl = false) {
        super(nodeType, parentForm, isTmpl);
        this._entries = [];
        this._observable = new Observable();
    }

    public get entries() {
        return this._entries;
    }

    public addEntry(e: SelectEntry) {
        this._entries.push(e);
        if (this._selectedEntry == undefined)
            this._selectedEntry = e;
    }

    public getValue() {
        return this._selectedEntry;
    }

    public setValue(v: SelectEntry) {
        if (this._selectedEntry !== v) {
            this._selectedEntry = v;
            this.notifyObservers({ "action": "select", "value": v });
        }
    }

    public get isValid(): boolean {
        return true;
    }

    public getLabel() {
        if (this._selectedEntry == undefined)
            return undefined;
        return this._selectedEntry.label;
    }

    protected verifyDependency(d: Dependency): boolean {
        switch (d.masterCondition.type) {
            case DependencyConditionType.HasValue:
                let mc: ValueDependencyCondition = <ValueDependencyCondition>d.masterCondition;
                return this._selectedEntry.value === mc.value;

            default:
                throw "SelectField.verifyDependency() : type de condition '" + DependencyConditionType[d.masterCondition.type] + "' non pris en charge";
        }
    }

    public updateLocalisation(loc: StringMap) {
        super.updateLocalisation(loc);

        for (let e of this._entries) {
            e.label = loc[e.value];
        }
    }

    /**
     * copie des membres
     */
    protected copyMembers(n: SelectField) {
        super.copyMembers(n);
        for (const e of this._entries)
            n.addEntry(e);
    }

    /**
     * crée une nouvelle instance
     */
    protected clone(): SelectField {
        return new SelectField(this.computeNodeType, this.parentForm);
    }

    public parseConfig(field: {}, data?: {}) {
        this._confId = field["id"];
        const values = field["select"];
        for (const v of values) {
            const e: SelectEntry = new SelectEntry(v["id"], undefined);
            this.addEntry(e);
        }
    }

    // interface IObservable

    /**
     * ajoute un observateur à la liste
     */
    public addObserver(o: Observer) {
        this._observable.addObserver(o);
    }

    /**
     * supprime un observateur de la liste
     */
    public removeObserver(o: Observer) {
        this._observable.removeObserver(o);
    }

    /**
     * notifie un événement aux observateurs
     */
    public notifyObservers(data: any) {
        this._observable.notifyObservers(data, this);
    }
}
