import { cLog, Result, ResultElement } from "jalhyd";

import { NgParameter } from "../formulaire/ngparam";
import { CalculatorResults } from "./calculator-results";

/**
 * type de graphe
 */
export enum GraphType {
    /**
     * histogramme
     */
    Histogram,

    /**
     * points indépendants reliés par une courbe
     */
    HistoLine,

    /**
     * graphe XY classique
     */
    Scatter,
}

export class FixedVarResults extends CalculatorResults {
    /**
     * résultats fixed (true) ou variés (false)
     */
    private _isFixed: boolean;

    /**
     * liste des résultats avec tous les paramètres fixés
     */
    private _fixedResults: Object[];

    /**
     * liste des résultats avec un paramètre variable
     */
    private _varResults: Object[];

    /**
     * titre de la 1ère colonne des résultats variés
     */
    private _variableParamHeader: string;

    /**
     * titre de la 2ème colonne des résultats variés
     */
    private _variableResultHeader: string;

    /**
     * titre du graphe des résultats variés
     */
    private _graphTitle: string;

    /**
     * type de graphe
     */
    public graphType: GraphType = GraphType.Scatter;

    /**
     * tableau de valeurs du graphe des résultats variés
     */
    private _varGraph: number[] = [];

    /**
     * journal de calcul
     */
    private _log: cLog;

    constructor() {
        super();
        this.reset();
    }

    public get log() {
        return this._log;
    }

    public get fixedResults() {
        return this._fixedResults;
    }

    public get varResults() {
        return this._varResults;
    }

    public get varGraph() {
        return this._varGraph;
    }

    public get isFixed() {
        return this._isFixed;
    }

    public reset() {
        this._isFixed = true;
        this._fixedResults = [];
        this._varResults = [];
        this._varGraph = [];
        this._variableParamHeader = undefined;
        this._variableResultHeader = undefined;
        this._graphTitle = undefined;
        this._log = new cLog();
    }

    public addFixedResultElement(p: NgParameter, v: ResultElement | number, displaySymbol: boolean) {
        if (v instanceof ResultElement)
            this._fixedResults.push({ "label": this.paramLabel(p, displaySymbol), "value": v });
        else if (typeof v === "number")
            this._fixedResults.push({ "label": this.paramLabel(p, displaySymbol), "value": new ResultElement(v) });
        else
            throw "appel FixedVarResults.addFixedResultElement() invalide : v = " + String(v);
    }

    public addVarResultElement(paramVal: number, resVal: ResultElement) {
        this._isFixed = false;
        this._varResults.push({ "param": new ResultElement(paramVal), "result": resVal });
        if (resVal.ok)
            this._varGraph.push(resVal.vCalc);
        else
            this._varGraph.push(0);
    }

    public addLogMessages(l: cLog) {
        for (let m of l.messages)
            this._log.add(m);
    }

    public get variableParamHeader() {
        return this._variableParamHeader;
    }

    public setVariableParamHeaderFromParameter(p: NgParameter, displaySymbol: boolean) {
        this._variableParamHeader = this.paramLabel(p, displaySymbol);
    }

    public get variableResultHeader() {
        return this._variableResultHeader;
    }

    public setVariableResultHeader(h: string) {
        this._variableResultHeader = h;
    }

    public setVariableResultHeaderFromParameter(p: NgParameter) {
        this._variableResultHeader = this.paramLabel(p, true);
    }

    public get graphTitle() {
        return this._graphTitle;
    }

    public set graphTitle(t: string) {
        this._graphTitle = t;
    }

    public get hasFixedResults(): boolean {
        return this._fixedResults.length > 0;
    }

    public get hasVarResults(): boolean {
        return this._varResults.length > 0;
    }

    public get hasResults(): boolean {
        return this.hasFixedResults || this.hasVarResults;
    }
}
